from typing import *

import wizwalker
from wizwalker.combat import CombatHandler
from wizwalker.combat import CombatMember
from wizwalker.combat.card import CombatCard

from .Util import safify, mem_safe
from .CombatConfigProvider import CombatConfigProvider, TargetType, TargetData, MoveConfig


class SprintyCombat(CombatHandler):
    def __init__(self, client: wizwalker.client.Client, config_provider: CombatConfigProvider):
        super().__init__(client)
        self.config = config_provider
        self.turn_adjust = 0
        self.cur_card_count = 0
        self.prev_card_count = 0
        self.was_pass = False
        self.had_first_round = False
        self.rel_round_offset = 0

    async def handle_combat(self):
        self.turn_adjust = 0
        self.cur_card_count = 0
        self.prev_card_count = 0
        self.rel_round_offset = 0
        self.was_pass = False
        self.had_first_round = False
        await super().handle_combat()

    @mem_safe
    async def get_members(self) -> List[CombatMember]:
        return await safify(super().get_members, timeout=self.config.memory_timeout)

    @mem_safe
    async def get_member_named(self, name: str) -> Optional[CombatMember]:
        members = await self.get_members()

        for member in members:
            if name == (await safify(member.name, timeout=self.config.memory_timeout)):
                return member
        return None

    @mem_safe
    async def get_member_vaguely_named(self, name: str) -> Optional[CombatMember]:
        members = await self.get_members()

        for member in members:
            if name.lower() in (await safify(member.name, timeout=self.config.memory_timeout)).lower():
                return member
        return None

    async def pass_button(self):
        self.was_pass = True
        await self.client.mouse_handler.click_window_with_name("Focus")

    @mem_safe
    async def get_cards(self) -> List[CombatCard]:  # extended to sort by enchanted
        cards = await safify(super().get_cards, timeout=self.config.memory_timeout)
        rese, res = [], []
        for card in cards:
            if await safify(card.is_enchanted, timeout=self.config.memory_timeout):
                rese.append(card)
            else:
                res.append(card)
        return rese + res

    @mem_safe
    async def get_card_named(self, name: str) -> Optional[CombatCard]:
        cards = await self.get_cards()

        for card in cards:
            if name == await card.name():
                return card

    @mem_safe
    async def get_card_vaguely_named(self, name: str) -> Optional[CombatCard]:
        cards = await self.get_cards()
        for card in cards:
            if name.lower() in (await safify(card.name, timeout=self.config.memory_timeout)).lower():
                return card

    @mem_safe
    async def get_card_counts(self) -> Tuple[int, int]:
        window = None
        while window is None:
            window, *_ = await self.client.root_window.get_windows_with_name("CountText")
        text: str = await safify(window.maybe_text, timeout=self.config.memory_timeout)
        _, count_text = text.splitlines()
        count_text = count_text[8:-9]
        count_text = count_text.replace("of", "").strip()  # I know this sucks
        res1, res2 = count_text.split()
        return int(res1), int(res2)

    @mem_safe
    async def get_castable_cards(self) -> List[CombatCard]:  # extension for castable cards only
        cards = await self.get_cards()
        res = []
        for card in cards:
            if await safify(card.is_castable, timeout=self.config.memory_timeout):
                res.append(card)
        return res

    @mem_safe
    async def get_castable_cards_named(self, name: str) -> List[CombatCard]:
        cards = await self.get_castable_cards()
        res = []

        for card in cards:
            if name == (await safify(card.name, timeout=self.config.memory_timeout)):
                res.append(card)

        return res

    @mem_safe
    async def get_castable_cards_vaguely_named(self, name: str) -> List[CombatCard]:
        cards = await self.get_castable_cards()
        res = []
        for card in cards:
            if name.lower() in (await safify(card.name, timeout=self.config.memory_timeout)).lower():
                res.append(card)

        return res

    @mem_safe
    async def get_castable_card_named(self, name: str) -> Optional[CombatCard]:  # extension to get only castable card
        cards = await self.get_castable_cards()

        for card in cards:
            if name == (await safify(card.name, timeout=self.config.memory_timeout)):
                return card

        return None

    @mem_safe
    async def get_castable_card_vaguely_named(self, name: str) -> Optional[CombatCard]:
        cards = await self.get_castable_cards()

        for card in cards:
            if name.lower() in (await safify(card.name, timeout=self.config.memory_timeout)).lower():
                return card

        return None

    @mem_safe
    async def get_castable_enchanted_card_named(self, name: str) -> Optional[CombatCard]:
        for s in await self.get_castable_cards_named(name):
            if await safify(s.is_enchanted, timeout=self.config.memory_timeout):
                return s
        return None

    @mem_safe
    async def get_castable_enchanted_card_vaguely_named(self, name: str) -> Optional[CombatCard]:
        for s in await self.get_castable_cards_vaguely_named(name):
            if await safify(s.is_enchanted, timeout=self.config.memory_timeout):
                return s
        return None

    @mem_safe
    async def get_boss_or_none(self) -> Optional[CombatMember]:
        for m in await self.get_members():
            if await safify(m.is_boss, timeout=self.config.memory_timeout):
                return m
        return None

    @mem_safe
    async def get_allies(self) -> List[CombatMember]:
        members = []
        my_client = await safify(self.get_client_member, timeout=self.config.memory_timeout)
        my_participant = await safify(my_client.get_participant, timeout=self.config.memory_timeout)
        my_team_id = await safify(my_participant.team_id, timeout=self.config.memory_timeout)
        my_id = await safify(my_participant.owner_id_full, timeout=self.config.memory_timeout)
        for mem in await self.get_members():
            participant = await safify(mem.get_participant, timeout=self.config.memory_timeout)
            if await safify(participant.team_id, timeout=self.config.memory_timeout) == my_team_id \
                    and await safify(participant.owner_id_full, timeout=self.config.memory_timeout) != my_id:
                members.append(mem)
        return members

    @mem_safe
    async def get_enemies(self) -> List[CombatMember]:
        members = []
        my_client = await safify(self.get_client_member, timeout=self.config.memory_timeout)
        my_participant = await safify(my_client.get_participant, timeout=self.config.memory_timeout)
        my_team_id = await safify(my_participant.team_id, timeout=self.config.memory_timeout)
        for mem in await self.get_members():
            participant = await safify(mem.get_participant, timeout=self.config.memory_timeout)
            if await safify(participant.team_id, timeout=self.config.memory_timeout) != my_team_id:
                members.append(mem)
        return members

    @mem_safe
    async def get_nth_ally_or_none(self, n: int) -> Optional[CombatMember]:
        allies = await self.get_allies()
        if len(allies) <= n:
            return None
        return allies[n]

    @mem_safe
    async def get_nth_enemy_or_none(self, n: int) -> Optional[CombatMember]:
        enemies = await self.get_enemies()
        if len(enemies) <= n:
            return None
        return enemies[n]

    async def try_get_config_target(self, target: TargetData) -> Union[bool, Optional[CombatMember]]:
        ttype = None
        data = None
        if target is not None:
            ttype = target.target_type
            data = target.extra_data
        else:
            return None

        if ttype is TargetType.type_boss:
            if boss := await self.get_boss_or_none():
                return boss
        elif ttype is TargetType.type_self:
            return await safify(self.get_client_member, timeout=self.config.memory_timeout)
        elif ttype is TargetType.type_aoe:
            return None
        elif ttype is TargetType.type_enemy:
            if data is None:
                if enemy := await self.get_nth_enemy_or_none(0):
                    return enemy
            else:
                if enemy := await self.get_nth_enemy_or_none(data):
                    return enemy
        elif ttype is TargetType.type_ally:
            if data is None:
                if ally := await self.get_nth_ally_or_none(0):
                    return ally
            else:
                if ally := await self.get_nth_ally_or_none(data):
                    return ally
        elif ttype is TargetType.type_named:
            if target.is_literal:
                if res := await self.get_member_named(data):
                    return res
            if res := await self.get_member_vaguely_named(data):
                return res

        return False

    async def try_execute_config(self, move_config: MoveConfig) -> bool:
        if move_config.move.card.name == "pass":
            await self.pass_button()
            return True

        target: Any
        spell = move_config.move

        target = await self.try_get_config_target(move_config.target)

        if target == False:  # Wouldn't want a None to mess it up
            return False

        to_cast = None
        if spell.card.is_literal:
            to_cast = await self.get_castable_enchanted_card_named(spell.card.name)
        else:
            to_cast = await self.get_castable_enchanted_card_vaguely_named(spell.card.name)
        if move_config.move.enchant is not None and move_config.move.enchant.name != "none" and to_cast is None:
            enchant_card = None
            if move_config.move.enchant.is_literal:
                enchant_card = await self.get_castable_card_named(move_config.move.enchant.name)
            else:
                enchant_card = await self.get_castable_card_vaguely_named(move_config.move.enchant.name)

            if enchant_card is None:
                return False
            to_enchant = None
            if spell.card.is_literal:
                to_enchant = await self.get_castable_card_named(spell.card.name)
            else:
                to_enchant = await self.get_castable_card_vaguely_named(spell.card.name)
            if to_enchant is None:
                return False
            await enchant_card.cast(to_enchant, sleep_time=self.config.cast_time * 2)
            if spell.card.is_literal:
                to_cast = await self.get_castable_enchanted_card_named(spell.card.name)
            else:
                to_cast = await self.get_castable_enchanted_card_vaguely_named(spell.card.name)
            self.cur_card_count -= 1
        elif to_cast is None:
            if spell.card.is_literal:
                to_cast = await self.get_castable_card_named(spell.card.name)
            else:
                to_cast = await self.get_castable_card_vaguely_named(spell.card.name)

        if to_cast is None:
            return False

        await to_cast.cast(target, sleep_time=self.config.cast_time)
        return True

    @mem_safe
    async def on_fizzle(self):
        self.turn_adjust -= 1

    async def handle_round(self):
        real_round = await self.round_number()
        self.cur_card_count = len(await self.get_cards()) + (await self.get_card_counts())[0]

        if not self.had_first_round:
            current_round = real_round - 1
            if current_round > 0:
                self.turn_adjust -= current_round
        else:
            if self.cur_card_count >= self.prev_card_count and not self.was_pass:
                await self.on_fizzle()

        self.was_pass = False
        current_round = (real_round - 1 + self.turn_adjust + self.rel_round_offset) % len(self.config.config.infinite_rounds)

        member = await safify(self.get_client_member, timeout=self.config.memory_timeout)
        if await safify(member.is_stunned, timeout=self.config.memory_timeout):
            self.turn_adjust -= 1
            await self.pass_button()
        else:
            round_config = self.config.get_real_round(real_round)
            if round_config is None:
                round_config = self.config.get_relative_round(current_round)
            else:
                self.rel_round_offset -= 1

            if round_config is not None:
                for p in round_config.priorities:  # go through rounds priorities
                    if await self.try_execute_config(p):
                        break  # we found a working priority and managed to cast it
                else:
                    await self.pass_button()
            else:  # Very bad. Probably using empty config
                raise RuntimeError(f"Full config fail! \"{self.config.filename}\" might be empty")

        self.had_first_round = True  # might go bad on throw
        self.prev_card_count = self.cur_card_count
