from typing import *

import asyncio

from .Exceptions import ExceptionalTimeoutError

def mem_safe(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


async def safify(fn: Callable[[], Coroutine[None, None, Any]], retry_time: float = 0.2,
                 timeout: float = 1.0, allow_none: bool = False):
    """
    :param fn: A async function that takes no parameters (use binding for functions with arguments)
    :param retry_time: Time between each attempt
    :param timeout: Time before the function is counted as failing
    :param allow_none: If the function is allowed to return None
    :return: The return value of the awaited coroutine (fn)
    """

    last_exception: List[Exception] = [Exception()]

    async def _inner():
        was_exception = False
        res = None
        while was_exception or (res is None and not allow_none):
            try:
                res = await fn()
                was_exception = False
            except Exception as ex:
                last_exception[0] = ex
                was_exception = True
                res = None
                await asyncio.sleep(retry_time)

        return res

    try:
        return await asyncio.wait_for(_inner(), timeout)
    except asyncio.TimeoutError:
        raise ExceptionalTimeoutError(f"Failed safe wrap of function.", last_exception[0])
